"""Globally used definitions"""
import itertools


# Generator of UID
_uid_gen = itertools.count(1)
uid_gen = lambda: next(_uid_gen)



def asp_value_to_pyvalue(value:str or iter or any) -> float or int:
    """

    >>> asp_value_to_pyvalue(2)
    2
    >>> asp_value_to_pyvalue('"3.4"')
    3.4
    >>> asp_value_to_pyvalue(('2', '"3.4"'))
    (2, 3.4)
    >>> asp_value_to_pyvalue(('2', ('"3.4"',))
    (2, (3.4,))

    """
    if isinstance(value, str):
        value = value.strip('"')
        if value.count('.') == 1 and all(part.isnumeric() for part in value.split('.')):
            return float(value)
        elif value.isnumeric():
            return int(value)
    if isinstance(value, (list, tuple)):
        return type(value)(map(asp_value_to_pyvalue, value))
    return value



"""Routines performing the compilation of input data toward the streams
of SpaceEngine scripts.

"""
import os
import itertools
from pprint import pprint

from . import serepr
from . import commons
from .commons import uid_gen
from .objects import OBJECTS, Orbit, Ring
from .objects_builder import ref


def compile_to_se(system_name:str, orbits:dict or tuple, objects:dict,
                  se_addons_dir:str or (str, str), overwrite:bool=False) -> (str):
    """Return names of written files"""
    star_lines, planet_lines = compile_to_gen(system_name, orbits, objects)
    system_name = system_name.replace(' ', '_')

    # choose the files to write
    if isinstance(se_addons_dir, str):
        star_file = se_addons_dir, 'addons/catalogs/stars/'
        planet_file = se_addons_dir, 'addons/catalogs/planets/'
        star_fname = planet_fname = system_name + '.sc'
    else:
        assert len(se_addons_dir) == 2, se_addons_dir
        star_file, planet_file = se_addons_dir
        star_fname = system_name + '.star.sc'
        planet_fname = system_name + '.planet.sc'
    star_file = os.path.expanduser(os.path.join(*star_file, star_fname))
    planet_file = os.path.expanduser(os.path.join(*planet_file, planet_fname))

    if not overwrite:
        if os.path.exists(star_file):
            raise ValueError("Star record already exists: {}".format(star_file))
        if os.path.exists(planet_file):
            raise ValueError("Planet record already exists: {}".format(star_file))
    with open(star_file, 'w') as fd:
        for line in star_lines:
            fd.write(line + '\n')
    with open(planet_file, 'w') as fd:
        for line in planet_lines:
            fd.write(line + '\n')
    return star_file, planet_file


def compile_to_gen(system_name:str, orbits:dict, objects:dict) -> ([str], [str]):
    """Return two generators of SpaceEngine script lines, one for the star file,
    one for the planet file.

    system_name -- name of the system in SpaceEngine
    orbits -- map from objects to orbiters and orbit info
    objects -- map from object uid to object definition

    """
    uids = {useruid: uid_gen() for useruid in objects.keys()}  # user uid -> uid
    orbits = tuple(uniformized_orbits(orbits))
    # print('ORBITS:', orbits)
    orbits, objects, old_uids = uidfy_data(orbits, objects, uids)
    # print('UIDFY:', orbits, objects, old_uids)
    name_of = lambda uid: system_name + '_' + str(type(objects[uid]).__name__).lower() + '_' + str(uid)

    # get root
    roots, inc_tree = make_inclusion_tree(orbits, uids)  # root uid -> {child_uid: {subchild uid: ...}}
    if len(roots) != 1:
        raise ValueError("Invalid number of roots. The {} roots are: {}".format(len(roots), ', '.join(map(str, roots))))
    root = next(iter(roots))
    del roots

    # compute se-repr of root
    star_lines = serepr.of_star_record(system_name)
    planet_lines = serepr.of_root(objects[root], name_of(root), system_name)
    # compute se-repr of each object
    for parent, child, orbit in orbits:
        lines = serepr.of_object(objects[child], name_of(child), name_of(parent),
                                 content=serepr.of_orbit(orbit))
        planet_lines = itertools.chain(planet_lines, lines)
    return star_lines, planet_lines


def make_inclusion_tree(orbits, uids) -> (set, dict):
    direct_inclusions = {}  # parent -> child
    for parent, child, orbit in orbits:
        direct_inclusions.setdefault(parent, set()).add(child)
    # print('DIRECT INCLUSIONS:', direct_inclusions)
    assert not any(parent in childs for parent, childs in direct_inclusions.items())
    def all_childs_of(parent, inclusions=direct_inclusions) -> [object]:
        """Yield all childs of given parent"""
        ret = []
        for child in inclusions.get(parent, ()):
            ret.append(child)
            ret.extend(all_childs_of(child, inclusions))
        return ret
    # close transitivity
    for parent in tuple(direct_inclusions.keys()):
        for child in all_childs_of(parent):
            direct_inclusions[parent].add(child)

    # print('INCLUSIONS:', direct_inclusions)
    roots = set(direct_inclusions.keys()) - set(itertools.chain.from_iterable(direct_inclusions.values()))
    if not roots:
        roots = set(direct_inclusions.keys())
    # print('ROOTS:', roots)
    return roots, direct_inclusions


def uidfy_data(orbits, objects, uids) -> (dict, dict, dict):
    """Return the same data, but with all user defined uid replaced by general uid,
    and all objects pushed into objects pool (no one-time-use objects).

    orbits -- orbit data
    objects -- map object uid -> object definition
    uids -- map user uid -> uid

    Returns:
    orbits -- orbit data
    objects -- map object uid -> object definition
    uids -- map uid -> old uid


    """
    new_orbits = []
    new_objects = {}
    old_to_new_uids = {}

    def make_uid(value):
        """Return an uid associated to given value. If value is not in uids map,
        then it's assumed to be a one-time-use lambda object"""
        already_seen = value in old_to_new_uids
        lambda_object = value not in uids and type(value) in OBJECTS
        raw_object = value not in uids and type(value) not in OBJECTS
        if already_seen:
            # print('EXISTING:', value, ' -> ', old_to_new_uids[value])
            return old_to_new_uids[value]
        if lambda_object:
            new_uid = uid_gen()
            new_objects[new_uid] = value
            # print('LAMBDA:', value, ' -> ', new_uid)
        elif raw_object:  # hope it's in refs
            # print('RAW LAMBDA:', value)
            return make_uid(ref(value))
        else:
            new_uid = uids[value]
            # print('USER DEFINED OBJECT:', value, objects[value], uids[value], new_uid)
            new_objects[new_uid] = objects[value]
            old_to_new_uids[value] = new_uid

        return new_uid

    # print('ORBITS:')
    # pprint(orbits)
    for parent, child, orbit in orbits:
        if isinstance(child, Ring):
            base_orbit = orbit._asdict()
            angle = orbit.angle
            for body, angle_step in zip(child.bodies, child.angle_steps):
                angle = (angle + angle_step) % 360
                base_orbit['angle'] = angle
                orbit = Orbit(**base_orbit)  # orbit specific to the child
                new_orbits.append((make_uid(parent), make_uid(body), orbit))
        else:
            new_orbits.append((make_uid(parent), make_uid(child), orbit))

    return tuple(new_orbits), new_objects, old_to_new_uids


def uniformized_orbits(orbits:dict or tuple) -> [tuple]:
    if isinstance(orbits, dict):
        for parent, childs in orbits.items():
            assert isinstance(childs, (tuple, list))
            if childs and isinstance(childs[0], (tuple, list)):
                assert len(childs[0]) == 2
                yield from ((parent, child, orbit) for child, orbit in childs)
            elif len(childs) == 2:
                yield (parent, *childs)
            else:
                raise ValueError("Unexpected orbit data for parent {}: {}".format(parent, childs))
    else:
        yield from orbits



"""Routines to build the objects."""

import inspect
from itertools import islice, cycle
from .objects import Planet, Orbit, Ring, Star, OBJECTS


orbit = Orbit
def ring(nb_body:int=None, body:object=None, *, angle_steps:float or iter=None) -> Ring:
    """Define a ring object.

    body -- either one object, or an iterable of objects
    nb_body -- number of body to put in the ring.

    If body is an iterable and nb_body is given, the ring will contains
    the nb_body firsts items of the iterable.

    """
    # invert arguments if any mixing
    if nb_body and not body and not isinstance(nb_body, int):
        body, nb_body = nb_body, None
    # compute bodies
    if inspect.isgenerator(body) or isinstance(body, (list, tuple)):
        if nb_body is None:
            bodies = tuple(body)
        else:  # take the given number in given iterable
            bodies = tuple(islice(cycle([body]), 0, int(nb_body)))
    else:
        bodies = (body,) * int(nb_body)
    nb_body = len(bodies)
    # compute angle steps
    if angle_steps is None:  # make them evenly distributed
        angle_steps = [360 / nb_body] * nb_body
    elif isinstance(angle_steps, (int, float)):
        angle_steps = [angle_steps] * nb_body
    else:
        angle_steps = tuple(islice(cycle(angle_steps), 0, nb_body))
    return Ring(
        nb_body=len(bodies),
        bodies=bodies,
        angle_steps=tuple(angle_steps),
    )


def star(spectral_class:str, solar_radius:float=None, solar_mass:float=None) -> Star:
    return Star(
        str(spectral_class),
        *_astre_mass_radius(solar_mass, solar_radius),
    )

def planet(cls:str, earth_radius:float=None, earth_mass:float=None) -> Planet:
    return Planet(
        str(cls),
        *_astre_mass_radius(earth_mass, earth_radius),
    )


def _astre_mass_radius(mass:float=None, radius:float=None) -> (float, float):
    if mass and radius is None:
        radius = mass
    elif radius and mass is None:
        mass = radius
    elif mass and radius:
        pass  # nothing to do
    else:
        raise ValueError("Astre definition: radius or mass must be given (received: {}, {}).".format(mass, radius))
    return float(mass), float(radius)


def structures_to_orbit() -> Orbit:
    pass


def ref(name:str, **kwargs) -> star or planet:
    if not isinstance(name, str):
        raise TypeError("Can't handle given value of type {} as name: {}".format(type(name), name))
    name = name.replace(' ', '_').lower()
    if name == 'sun':
        payload = star, {'solar_mass': 1, 'solar_radius': 1, 'spectral_class': 'G2V'}
    elif name == 'red_dwarf':
        payload = star, {'solar_mass': 0.1, 'spectral_class': 'M5V'}
    elif name == 'blue_giant':
        payload = star, {'solar_mass': 10, 'spectral_class': 'O9'}
    elif name == 'earth':
        payload = planet, {'cls': 'terra', 'earth_mass': 1, 'earth_radius': 1}
    elif name == 'moon':
        payload = planet, {'cls': 'luna', 'earth_mass': 0.1}
    elif name == 'black_hole':
        payload = star, {'spectral_class': 'X', 'solar_mass': 2000}
    else:
        raise NotImplementedError("Non-implemented name: '{}'".format(name))
    func, fixed_kwargs = payload
    return func(**{**fixed_kwargs, **kwargs})



"""Functions bridging internal data and SpaceEngine representation.

"""

def se_star(parent:str, name:str, spectral_class:str=None, solar_masses:float=None, solar_radius:float=None, procgen:bool=False, content:[str]=(), accretion_disk:bool=False) -> [str]:
    yield 'Star    "{}"'.format(name)
    yield '{'
    yield '    ParentBody "{}"'.format(parent)
    if spectral_class:
        yield '    Class      "{}"'.format(spectral_class)
    if solar_masses:
        yield '    MassSol    {}'.format(solar_masses)
    if solar_radius:
        yield '    RadSol    {}'.format(solar_radius)
    if not procgen:
        yield '    NoPlanets  true'
    if not accretion_disk:
        yield '    AccretionDisk no'
    if content:
        yield from content
    yield '}'

def se_planet(parent:str, name:str, cls:str=None, earth_masses:float=None, earth_radius:float=None, procgen:bool=False, content:[str]=()) -> [str]:
    yield 'Planet    "{}"'.format(name)
    yield '{'
    yield '    ParentBody     "{}"'.format(parent)
    if cls:
        yield '    Class          "{}"'.format(cls)
    if earth_masses:
        yield '    Mass           {}'.format(earth_masses)
    if earth_radius:
        yield '    Radius         {}'.format(round(float(earth_radius) * 6378, 2))
    if not procgen:
        yield '    NoPlanets  true'
    if content:
        yield from content
    yield '}'


SE_ORBIT = """
    Obliquity       {obliquity}
    Orbit
    {{
        RefPlane        "Equator"
        SemiMajorAxis   {semimajoraxis}
        Eccentricity    {eccentricity}
        Inclination     {inclination}
        MeanAnomaly     {angle}
        AscendingNode   0
    }}
""".strip('\n')


def se_orbit(semimajoraxis:float, eccentricity:float=0, obliquity:float=0, inclination:float=0, angle:float=0, retrograde:bool=False) -> [str]:
    yield SE_ORBIT.format(semimajoraxis=semimajoraxis, eccentricity=eccentricity,
                          inclination=inclination + (180 if retrograde else 0),
                          angle=angle, obliquity=obliquity)



"""Routines to build model from ASP data"""

import clyngor
from .objects import Model, Orbit, ORBIT_ARGS_ORDER, STAR_ARGS_ORDER, PLANET_ARGS_ORDER
from .objects_builder import ref
from .commons import uid_gen, asp_value_to_pyvalue


def data(fname:str) -> [dict]:
    """Yield the extracted data from given file"""
    yield from clyngor.solve(fname).by_predicate.parse_args.careful_parsing


def root_info(asp_model, objects:dict) -> (str, str, ...):
    """Return the root uid, the system name, and data to be given
    to populate_orbits function.

    asp_model -- dictionnary containing definition of asp atoms
    objects -- dict. Modified, so root_uid maps to the root definition.

    """
    roots = tuple(args for args in asp_model.get('root', ()) if len(args) >= 2)
    if len(roots) > 1:
        print('Multiple roots. Bad.')
        exit(1)
    elif len(roots) == 1:
        format_ = lambda x: (str(x).strip('"') if isinstance(x, str) else x)
        args = tuple(map(format_, roots[0]))
        if len(args) == 2:
            root, name = args
            uid = name
        elif len(args) == 3:
            root, name, uid = args
        else:
            raise ValueError("Expect a root/2 or root/3 atom.")
    else:
        print('No root. Bad.')
        exit(1)  # TODO: determine the root by yourself

    types, names = _types_and_names(asp_model, objects)
    root_type, NAME, UID = types.get(root, root), name, uid
    root_uid = UID + '__' + NAME
    system_name = root_uid + ' system'
    objects[root_uid] = ref(root_type)
    return root_uid, system_name, types, names, root


def populate_orbits(asp_model, orbits:list, objects:dict, root_uid:str,
                    types:dict, names:dict, root_number:int):
    names[root_number] = root_uid
    for args in asp_model.get('orbit', ()):
        orbits.extend(gen_orbits(args, objects, types, names))


def gen_orbits(args:tuple, objects:dict, types:dict, names:dict) -> [Orbit]:
    if len(args) >= 3:
        parent, child, orbit_params, *properties = args

        # Compute orbit parameters
        if isinstance(orbit_params, (str, int)):
            orbit_params_map = {'semimajoraxis': asp_value_to_pyvalue(orbit_params)}
        elif not isinstance(orbit_params, tuple) or len(orbit_params) != 2:
            raise ValueError("Unexpected orbit parameters: {}".format(orbit_params))
        else:  # it's a tuple !
            predicat, orbit_params = orbit_params
            orbit_params_map = {
                k: asp_value_to_pyvalue(v)
                for k, v in dict(zip(ORBIT_ARGS_ORDER, orbit_params)).items()
            }
        distance = orbit_params_map['semimajoraxis']  # shortcut

        # Compute objects representation
        if isinstance(child, tuple) and child[0] in {'planet', 'star'}:
            child, obj = _asp_tuple_to_object(child)
            objects[child] = obj
        if isinstance(parent, tuple) and parent[0] in {'planet', 'star'}:
            parent, obj = _asp_tuple_to_object(parent)
            objects[parent] = obj

        properties = set(properties)
        retrograde = 'retrograde' in map(lambda s: s.strip('"'), properties)
        parent = names.get(parent, parent)
        if isinstance(child, (int, str)):
            try:
                child_type = types.get(child, child)
            except KeyError:
                raise ValueError("Unknow type (is/2) for element of uid {}".format(child))
            child = names.setdefault(child, str(child) + str(uid_gen()))
        # orbit cases
        if isinstance(child, tuple) and child[0] == 'orbit':
            # recursive handling of the childs
            subargs = child[1]
            assert len(subargs) >= 3
            # link between parent and direct child
            yield from gen_orbits((parent, subargs[0], distance), objects, types, names)
            # subsequent links
            yield from gen_orbits(subargs, objects, types, names)
        elif isinstance(child, tuple) and child[0] == 'ring':
            ring_params = child[1]
            if 'angle' in orbit_params_map:
                raise ValueError("Orbit parameters specify an angle, but the object is a ring.")
            assert len(child) == 2 and len(ring_params) >= 2
            ring_type, number, *properties = ring_params
            angle_step = 360 / number
            ring_orbits = []
            for idx in range(number):
                angle = idx * angle_step
                orbit = Orbit(**orbit_params_map, angle=angle, retrograde=retrograde)
                child_uid = ring_type + str(uid_gen())
                objects[child_uid] = ref(ring_type)
                ring_orbits.append((names.get(parent, parent), child_uid, orbit))
            for prop in properties:
                if len(prop) == 2 and prop[0] == 'orbit' and len(prop[1]) >= 3:
                    subargs = prop[1]
                    index = subargs[0]
                    if (index-1) in range(len(ring_orbits)):
                        parent = ring_orbits[index][1]
                        yield from gen_orbits((parent, *subargs[1:]), objects, types, names)
            yield from ring_orbits
        elif isinstance(child, str):
            objects[child] = ref(child_type)
            orbit = Orbit(retrograde=retrograde, **orbit_params_map)
            yield names.get(parent, parent), child, orbit
        else:
            raise ValueError("Non handlable orbit data '{}' (understood as '{}')"
                             "".format(args, (parent, child, distance, properties)))
    else:
        print('Non valid: {}'.format(args))


def _types_and_names(asp_model, objects:dict) -> (dict, dict):
    types = {}  # integer uid -> object types
    names = {}  # integer uid -> object name
    for args in asp_model.get('is', ()):
        if len(args) == 2:
            uid, obj = args
            if uid in types:
                raise ValueError("UID {} given multiple (at least two) times. "
                                 "One for '{}', One for '{}'"
                                 "".format(uid, types[uid], obj))
            if isinstance(obj, tuple):
                obj, obj_repr = _asp_tuple_to_object(obj)
                objects[obj] = obj_repr
            types[uid] = obj
            names[uid] = obj + str(uid_gen())
    return types, names

def _asp_tuple_to_object(atom:tuple) -> (str, dict):
    """Return the unique id and the dict representation of given planet or star object"""
    assert isinstance(atom, tuple) and atom[0] in {'planet', 'star'}
    args_order = PLANET_ARGS_ORDER if atom[0] == 'planet' else STAR_ARGS_ORDER
    new_uid = str(uid_gen())
    data = dict(zip(args_order, atom[1]))
    data['type'] = atom[0]
    return new_uid, data



"""Definitions for general objects"""


from collections import namedtuple
from . import templates


ORBIT_ARGS_ORDER = ('semimajoraxis', 'eccentricity', 'obliquity',
                         'inclination', 'angle', 'retrograde')
STAR_ARGS_ORDER = ('spectral_class', 'solar_mass', 'solar_radius')
PLANET_ARGS_ORDER = ('cls', 'earth_mass', 'earth_radius')
RING_ARGS_ORDER = ('nb_body', 'bodies', 'angle_steps')


# Model structure
Model = namedtuple('Model', 'system_name, orbits, objects')
# Astres structures
Star = namedtuple('Star', STAR_ARGS_ORDER)
Planet = namedtuple('Planet', PLANET_ARGS_ORDER)
Ring = namedtuple('Ring', RING_ARGS_ORDER)
# Orbit structures
Orbit = namedtuple('Orbit', ORBIT_ARGS_ORDER)
Orbit.__new__.__defaults__ = 0, 0, 0, 0, False

OBJECTS = {Star, Planet, Ring, Orbit}


def se_repr(template:callable, framed=True):
    if framed:
        def se_repr_gen(obj:object, name:str, parent:str, content:[str]=()) -> [str]:
            yield from template(parent, name, *obj, content=content)
    else:
        def se_repr_gen(obj:object) -> [str]:
            yield from template(*obj)
    return se_repr_gen
Orbit.se_repr = se_repr(templates.se_orbit, framed=False)
Planet.se_repr = se_repr(templates.se_planet)
Star.se_repr = se_repr(templates.se_star)



"""CLI entry point for the package.

"""

import os
from . import cli

from . import get_models, models_to_se


if __name__ == '__main__':
    args = cli.parse_args(description=__doc__)
    # print('ARGS:', args)
    if args.se_dir is None:  # user choose a test dir
        outdir = args.test_dir, args.test_dir
    else:  # user gives us a good ol' space engine directory
        outdir = args.se_dir
    models = get_models(args.infile)
    models_to_se(models, outdir)



"""Routines to build model from JSON data"""

import json
from .objects import Model, Orbit
from .objects_builder import ref
from .commons import uid_gen


def data(fname:str) -> [dict]:
    """Yield the extracted data from given file"""
    with open(fname) as fd:
        data = json.load(fd)
        if isinstance(data, list):
            yield from (datum for datum in data if isinstance(datum, dict))
        else:
            assert isinstance(data, dict)
            yield data


def root_info(json_data:dict, objects:dict) -> (str, str):
    """Return the root uid and the system name.

    json_data -- json dict containing data
    objects -- dict. Modified, so root_uid maps to the root definition.

    """
    root_type, NAME, UID = json_data['type'], json_data['name'], json_data.get('UID', json_data['name'])
    root_uid = UID + '__' + NAME
    system_name = root_uid + ' system'
    objects[root_uid] = ref(root_type)
    return root_uid, system_name


def populate_orbits(json_data:dict, orbits:list, objects:dict, root_uid:str):
    assert isinstance(json_data, dict), json_data
    if 'childs' in json_data or 'child' in json_data:
        iter_childs = json_data.get('childs', json_data.get('child'))
        if isinstance(iter_childs, dict): iter_childs = [iter_childs]
        for child in iter_childs:
            new_orbits = list(gen_orbits(root_uid, child, objects))
            for subchild in child.get('childs', ()):
                new_orbits += gen_orbits(child['UID'], subchild, objects)
            if 'child' in child:
                new_orbits += gen_orbits(child['UID'], child['child'], objects)
            # handle childs of specific stars of rings
            if 'childof' in child and isinstance(child['childof'], dict):
                for parent_index, childs in child['childof'].items():
                    if isinstance(childs, dict):  # one child
                        childs = [childs]
                    elif isinstance(childs, list):  # list of child
                        pass
                    else:
                        raise ValueError("Unhandlable json value for ring child: {}".format(childs))
                    if not parent_index.isnumeric():
                        raise ValueError("Parent index for childs of rings element must be a integer value, not '{}'".format(parent_index))
                    parent_uid = new_orbits[int(parent_index)][1]
                    for subchild in childs:
                        new_orbits += gen_orbits(parent_uid, subchild, objects)

            orbits.extend(new_orbits)


def gen_orbits(parent:str, child:dict, objects:dict) -> [Orbit]:
    retrograde = child.get('retrograde', False)
    child_type = child.get('type')
    distance = child['distance']
    if isinstance(child_type, (list, tuple)) and child_type[0] == 'ring':
        ring_params = child_type[1:]
        assert len(child_type) >= 2 and len(ring_params) == 2
        number, ring_type = ring_params
        angle_step = 360 / number
        for idx in range(number):
            angle = idx * angle_step
            new_uid = ring_type + str(uid_gen())
            child['UID'] = new_uid
            objects[new_uid] = ref(ring_type)
            yield parent, new_uid, Orbit(distance, angle=angle, retrograde=retrograde)
    elif isinstance(child_type, str):
        new_uid = child_type + str(uid_gen())
        child['UID'] = new_uid
        objects[new_uid] = ref(child_type)
        yield parent, new_uid, Orbit(distance, retrograde=retrograde)



"""Routines for model manipulation.

"""

import os
from . import asp_model
from . import json_model
from . import commons
from .objects import Model
from .compile import compile_to_gen, compile_to_se
from .objects_builder import planet, orbit, ring, star, ref


def models_to_se(models:[Model] or Model, se_addons_dir:str or (str, str)):
    """Print given models into their dedicated files, into se addon dir"""
    for model in ([models] if isinstance(models, Model) else models):
        print('SYSTEM: {} ({} orbits)'.format(model.system_name, len(model.orbits)))
        print('FILES:', '\n       '.join(model_to_se(model, se_addons_dir)))
        print()

def model_to_se(model:Model, se_addons_dir:str):
    """Print given model into its dedicated files, into se addon dir"""
    return compile_to_se(model.system_name, model.orbits, model.objects,
                         se_addons_dir, overwrite=True)


def get_models(fname:str or iter) -> [Model]:
    """Yield Models from file of given name, either in JSON or ASP"""
    ext_to_class = {
        '.lp': asp_model,
        '.json': json_model,
    }
    ext = os.path.splitext(fname)[1]
    try:
        extractor = ext_to_class[ext]
    except KeyError:
        raise ValueError("Extension {} is not handled".format(ext))
    yield from (model_from(datum, extract=extractor) for datum in extractor.data(fname))


def model_from(asp_model, extract) -> Model:
    """Extract Model from ASP answer sets"""
    orbits = list()
    objects = {}

    root_uid, system_name, *other_data = extract.root_info(asp_model, objects)
    assert root_uid in objects

    extract.populate_orbits(asp_model, orbits, objects, root_uid, *other_data)  # side effect

    return Model(str(system_name), tuple(orbits), objects)



"""Templating of SpaceEngine script for various objects.

"""
from . import templates


def of_star_record(system_name:str) -> [str]:
    yield 'Remove "' + system_name + '"'
    yield 'StarBarycenter    "' + system_name + '"'
    yield '{}'

def of_root(root:object, name:str, system_name:str):
    """SpaceEngine script representation of the root."""
    return root.se_repr(name, system_name)

def of_object(obj:object, name:str, parent:str, *, content:[str]=()):
    """SpaceEngine script representation of given star or planet."""
    return obj.se_repr(name, parent, content=content)

def of_orbit(orbit:object):
    """SpaceEngine script representation of given Orbit."""
    return orbit.se_repr()



"""Definition of the CLI.
"""

import os
import argparse


def parse_args(description:str, args:iter=None) -> dict:
    return cli_parser(description).parse_args(args)

def existant_file(filepath:str) -> str:
    """Argparse type, raising an error if given file does not exists"""
    if not os.path.exists(filepath):
        raise argparse.ArgumentTypeError("file {} doesn't exists".format(filepath))
    return filepath

def existant_dir(filepath:str) -> str:
    """Argparse type, raising an error if given file does not exists"""
    if not os.path.isdir(filepath):
        raise argparse.ArgumentTypeError("directory {} doesn't exists".format(filepath))
    return filepath

def se_dir(filepath:str) -> str:
    """Argparse type, raising an error if given directory is not a SpaceEngine directory"""
    existant_dir(filepath)
    dirs = set()
    expected_dirs = {'addons', 'config', 'system'}
    dirs = {entry.name for entry in os.scandir(filepath) if entry.is_dir()}
    not_found = expected_dirs - dirs
    if not_found:
        raise argparse.ArgumentTypeError("SpaceEngine directory {} doesn't "
                                         "contains expected subdir: {}"
                                         "".format(filepath, ', '.join(no_found)))
    return filepath

def writable_file(filepath:str) -> str:
    """Argparse type, raising an error if given file is not writable.
    Will delete the file !
    """
    try:
        with open(filepath, 'w') as fd:
            pass
        os.remove(filepath)
        return filepath
    except (PermissionError, IOError):
        raise argparse.ArgumentTypeError("file {} is not writable.".format(filepath))


def cli_parser(description:str) -> argparse.ArgumentParser:
    # main parser
    parser = argparse.ArgumentParser(description=description.strip())

    parser.add_argument('infile', type=existant_file,
                        help="Name of the input file containing the system to render")

    # user may want to put automatically the system in place, or anywhere without treatment
    outdir = parser.add_mutually_exclusive_group()
    outdir.add_argument('--se-dir', '-o', type=se_dir,
                        help="Where SpaceEngine is installed")
    outdir.add_argument('--test-dir', '-t', type=existant_dir, default='.',
                        help="Where to put the files")

    parser.add_argument('--overwrite', action='store_true',
                        help="Rewrite existing files")

    return parser
